/*
  CPP_CONTEST=2013
  CPP_PROBLEM=B
  CPP_LANG=C+OPENMP+MPI
  CPP_PROCESSES_PER_NODE=marte 6 mercurio 6
*/

/* RECORD
Youssef Said Khloufi 
students of the University of Murcia, Spain
records established December 4th, 2013
time 1810 msec
speed-up 16.3
*/ 

/*The  rectangles to be updated are divided between the processes,
and the local matrices are accumulated in process 0.
OpenMP threads are used in the updating of each rectangle.
Balancing of the rectangles to be updated could be considered,
together with higher level parallelization in the local use of OpenMP.*/

#include <stdlib.h>
#include "mpi.h"
#include <omp.h>
#include <math.h>
#include <stdio.h>
 
void accumulatelocal(int *a,int N,int v,int r1,int c1,int r2,int c2)
{
  int i,j;
  int *pos;

  for(i=r1;i<=r2;i++)
  {
    pos=&a[i*N];
    for(j=c1;j<=c2;j++)
    {
      pos[j]+=v;
    }
  }
}
    
//void  sec(int N,int M,int *a,int *value,int *r1,int *c1,int *r2,int *c2,int nodo,int np)
void  sec(int N,int M,int *a,int *value,int *r1,int *c1,int *r2,int *c2)
{
  int rank,dimProc;
  int i_ini,i_fin,j_ini, j_fin;
  int i,p;
  // M = 300.000 como maximo ( * 5 * 4 bytes)
  char buffer[6000010];  // 6 MB aprox
  int posicion, packsize;
    
  MPI_Status status;
  MPI_Comm_rank(MPI_COMM_WORLD,&rank);
  MPI_Comm_size(MPI_COMM_WORLD,&p);
    

  // distribuir N, M y La matriz "a" a todos los procesos 
  if(rank==0){
    packsize = 0;
    MPI_Pack(&N,1,MPI_INT,buffer,6000010,&packsize,MPI_COMM_WORLD);
    MPI_Pack(&M,1,MPI_INT,buffer,6000010,&packsize,MPI_COMM_WORLD);
    MPI_Pack(value,M,MPI_INT,buffer,6000010,&packsize,MPI_COMM_WORLD);
    MPI_Pack(r1,M,MPI_INT,buffer,6000010,&packsize,MPI_COMM_WORLD);
    MPI_Pack(c1,M,MPI_INT,buffer,6000010,&packsize,MPI_COMM_WORLD);
    MPI_Pack(r2,M,MPI_INT,buffer,6000010,&packsize,MPI_COMM_WORLD);
    MPI_Pack(c2,M,MPI_INT,buffer,6000010,&packsize,MPI_COMM_WORLD);    
    
    dimProc = N / p; // Zona de la matriz de N * dimProc para cada Nodo
    
    // Determinar zona de la matriz que el proceso tiene que calcular
    i_ini = 0;
    i_fin = i_ini+dimProc-1;    
  }
  // Es necesario mandar antes el tamaño de los elementos el buffer
  MPI_Bcast( &packsize, 1, MPI_INT, 0, MPI_COMM_WORLD );
  // Distribuir el buffer de empaquetamiento a todos los procesos
  MPI_Bcast(buffer, packsize, MPI_PACKED, 0, MPI_COMM_WORLD);
  
  if (rank != 0){
    posicion=0;
    MPI_Unpack(buffer, packsize, &posicion,&N,1,MPI_INT,MPI_COMM_WORLD);
    MPI_Unpack(buffer, packsize, &posicion,&M,1,MPI_INT,MPI_COMM_WORLD);       
    
    // Reservo memoria dinamica para la matriz "a"
    a = (int *) calloc(sizeof(int),N*N);
    //reservo espacio para los vectores value,r1,c1,r2,c2
    
    value = (int *) malloc(sizeof(int)*M);
    r1=(int *) malloc(sizeof(int)*M);
    c1=(int *) malloc(sizeof(int)*M);
    r2=(int *) malloc(sizeof(int)*M);
    c2=(int *) malloc(sizeof(int)*M);
        
    MPI_Unpack(buffer, packsize, &posicion,value,M,MPI_INT,MPI_COMM_WORLD);
    MPI_Unpack(buffer, packsize, &posicion,r1,M,MPI_INT,MPI_COMM_WORLD);
    MPI_Unpack(buffer, packsize, &posicion,c1,M,MPI_INT,MPI_COMM_WORLD);
    MPI_Unpack(buffer, packsize, &posicion,r2,M,MPI_INT,MPI_COMM_WORLD);
    MPI_Unpack(buffer, packsize, &posicion,c2,M,MPI_INT,MPI_COMM_WORLD);
    
    dimProc = N / p; // Zona de la matriz de N * dimProc para cada Nodo

    // Determinar zona de la matriz que el proceso tiene que calcular
    i_ini = (rank * dimProc);
    if(rank==p-1) // el ultimo
      i_fin = N-1;
    else
      i_fin = i_ini+dimProc-1;
  }
  
  // Todos recorren filas enteras (division por fila)
  j_ini = 0; 
  j_fin = N-1;

  omp_set_num_threads(11);
  
  #pragma omp parallel private(i)
  {
    int np,iam,dim;
    int i_ini_thread,i_fin_thread,j_ini_thread, j_fin_thread;
    
    np = omp_get_num_threads();
    iam = omp_get_thread_num();    
      
    dim = (i_fin-i_ini+1) / np; // Zona de la submatriz a utilizar
    
    if(iam==0)
      i_ini_thread = i_ini; // donde empieza la region del proceso
    else
      i_ini_thread = i_ini + (iam * dim);
    if(iam==np-1) // el ultimo hilo
      i_fin_thread = i_fin; // donde termina la region del proceso
    else
      i_fin_thread = i_ini_thread+dim-1;
      
    j_ini_thread = 0;
    j_fin_thread = N-1;
    
    for(i=0;i<M;i++)
    {    
      int fila_ini,fila_fin,col_ini,col_fin; // indices de la submatriz de interseccion

      // Averiguar interseccion
        
      if ((r2[i] < i_ini_thread) || // a la izquierda del otro
        (r1[i] > i_fin_thread) || // a la derecha del otro
        (c2[i] < j_ini_thread) || // encima del otro
        (c1[i] > j_fin_thread)) // abajo del otro
        // No hacer nada porque no hay interseccion
        ; 
      else{
        // calcular indices de interseccion
        if(r1[i]>i_ini_thread)
          fila_ini = r1[i];
        else
          fila_ini = i_ini_thread;
        if(r2[i]<i_fin_thread)
          fila_fin = r2[i];
        else
          fila_fin = i_fin_thread;
        if(c1[i]>j_ini_thread)
          col_ini = c1[i];
        else
          col_ini = j_ini_thread;
        if(c2[i]<j_fin_thread)
          col_fin = c2[i];
        else
          col_fin = j_fin_thread;
                
        accumulatelocal(a,N,value[i],fila_ini,col_ini,fila_fin,col_fin);
      }
    }
  }

  // Aqui se mezclan los hilos en un proeso unico     
   
  if(rank!=0){ // Envio el trozo calculado al nodo raiz
    int size = (N*(i_fin-i_ini+1)); // num filas tratadas * elementos en cada fila
    MPI_Send(&a[i_ini*N],size,MPI_INT,0,rank, MPI_COMM_WORLD);
    free(a);
    free(value);
    free(r1);
    free(c1);
    free(r2);
    free(c2);
  }
  else{ // Si soy el nodo raiz recibo los trozos calculados de los procesos
    int max = (N*N) -( dimProc * N * (p-1) ); // el tamaño del trozo mas grande
    int donde;    
    for(i=1; i < p; i++){
      donde = i*dimProc*N;
      MPI_Recv(&a[donde],max,MPI_INT,i,i,MPI_COMM_WORLD,&status);
    }
  } 
}
